#include <httpd.h>
#include <http_log.h>

#include "config.h"
#include "wasap.h"


/*
	he version of an HTTP message is indicated by an HTTP-Version field
	in the first line of the message.
	
  	HTTP-Version   = "HTTP" "/" 1*DIGIT "." 1*DIGIT
  
	Note that the major and minor numbers MUST be treated as separate
	integers and that each MAY be incremented higher than a single
	digit. Thus, HTTP/2.4 is a lower version than HTTP/2.13, which in
	turn is lower than HTTP/12.3. Leading zeros MUST be ignored by
	recipients and MUST NOT be sent.
		
*/

#define LOG_ERROR(wr,level,fmt, args...) \
	 ap_log_rerror(APLOG_MARK,level,wr->pool,wr->rec,fmt, ##args)

int check_proto(wasap_rec * wr, char *protocol)
{
	char *ptr;
	int major, minor;
	
	
	if (strncmp(protocol, "HTTP/", 5)) {
		LOG_ERROR(wr,APLOG_CRIT,
			"mod_wasap: illegal protocol string [%s]",
			ap_escape_logitem(wr->pool,protocol));
		
		exec_alarm(wr->rec,"illegal protocol string");
		return (FALSE);
	}
	
	sscanf(protocol, "HTTP/%d.%d", &major, &minor);
	
	if (MAKE_VER(major, minor) > MAX_VER) {
		LOG_ERROR(wr,APLOG_CRIT,
				"mod_wasap: illegal protocol number: %d.%d",
				major, minor);
		
		exec_alarm(wr->rec,"illegal protocol number");
		return (FALSE);
	}
	
	return (TRUE);
	
}


int check_scheme(wasap_rec *wr, char *request)
{
	char *ptr, *ptr2, c;
	unsigned long x;
	
	
	ptr = strchr(request,' ');
	if (ptr == NULL)
		return(FALSE);
	
	while(*ptr == ' ')
		ptr++;
	
	
	ptr2 = ptr;

	if (*ptr == '/')
		return (TRUE);
	
	if ((ptr = strchr(ptr, '/')) == NULL) {
		LOG_ERROR(wr,APLOG_CRIT,
			      "mod_wasap: invalid scheme [%s]",
			      ap_escape_logitem(wr->pool,request));
		
		exec_alarm(wr->rec,"invalid scheme");
		return (FALSE);
	}
	
	if (*(ptr - 1) != ':' || *(ptr + 1) != '/') {
		LOG_ERROR(wr,APLOG_CRIT,
				"mod_wasap: illegal scheme syntax in [%s]",
				ap_escape_logitem(wr->pool,request));
		
		exec_alarm(wr->rec,"invalid scheme syntax");
		return (FALSE);
	}

	ptr += 2;
	if (strncmp(ptr2, "http://", ptr - ptr2)) {
		LOG_ERROR(wr, APLOG_CRIT,
			      "mod_wasap: illegal scheme type in [%s]",
			      ap_escape_logitem(wr->pool,request));
		
		exec_alarm(wr->rec,"illegal scheme type");
		return (FALSE);
	}

	if (strncasecmp(SERVER_NAME, ptr, strlen(SERVER_NAME)) != 0) {
		LOG_ERROR(wr, APLOG_CRIT,
			      "illegal server: %s", ap_escape_logitem(wr->pool,ptr));
		
		exec_alarm(wr->rec,"illegal server");
		return (FALSE);
	}
	
	ptr += strlen(SERVER_NAME);
	
	if (*ptr == ':') {
		ptr++;

		x = strtol(ptr, (char **) NULL, 10);
		if (errno == ERANGE) {
			LOG_ERROR(wr, APLOG_CRIT,
				"mod_wasap: port number out of range: [%s]",
				ap_escape_logitem(wr->pool,ptr));
			
			exec_alarm(wr->rec,"illegal port number");
			return (FALSE);
		}
		
		if (x > 65535 || x < 1) {
			LOG_ERROR(wr, APLOG_CRIT,
				      "mod_wasap: scheme has illegal port[%u]",
				      x);
			
			exec_alarm(wr->rec,"illegal port number");
			return (FALSE);
		}
		
		
		/* missing call to trans_uri() */
		
		for (; *ptr && *ptr != '/'; ptr++)
			if (*ptr > 0x39 || *ptr < 0x30) {
				LOG_ERROR(wr, APLOG_CRIT,
					      "mod_wasap: illegal char in uri: [%c] in [%s]",
					      *ptr, ap_escape_logitem(wr->pool,request));
				
				exec_alarm(wr->rec,"illegal character in uri");
				return (FALSE);
			}
	}
	
	return (TRUE);
	
}


static int trans_char(char *str, char *buf)
{
	unsigned char *p, *s;
	unsigned char tmp, mul;
	/*
	unsigned char *ucd[6] = { 0xc0, 0xe0, 0xf0, 0xf8, 0xfc };
	unsigned char ucnum=0, uccnt=0;
	*/
	
	for (p = str, s = buf; *p;) {
		if (*p == '%') {
			p++;
			
			/* 
			   Use a bit shift to first multiply the X part of %Xx
			   by 16, making it the H.O. four bits.
			   Later multiply the x by 1 to make it the L.O.
			   four bits, forming our byte.
			   
			   We dont try to translate anything besides 0-9,A-F,a-f
			   (c0z its illegal, you silly hax0r)
			   
			   (should later be done ENTIRELY with bit shifts)
			 */
			
			
			for (tmp = 0, mul = 16; mul > 0 && *p; mul >>= 4, p++) {
				
				if (*p >= '0' && *p <= '9')
					tmp += (*p - 0x30) * mul;
				
				else if (*p >= 'A' && *p <= 'F')
					tmp += (*p - 0x37) * mul;
				
				else if (*p >= 'a' && *p <= 'f')
					tmp += (*p - 0x57) * mul;
				
				else
					return (0);
				
			}
			
			/* % with one char might be able to evade the system
			 * somehow...
			 */
			
			if (tmp == 0)
				return(FALSE);
			
			/*
			if (tmp == 0x80)
				ucdcnt--;
			} else {
				if (ucdcnt==1 && tmp == 0x80)
						return(FALSE);
				
				for (ucdnum=0; ucdnum<sizeof(ucd)/sizeof(char); ucdnum++) {
					if (ucd[ucdnum] == tmp)
						ucdcnt = ucdnum+1;
				}
				
			}
			*/
			
			/*
			if (!isprint(tmp))
				return(FALSE);
			*/
			
			*s++ = tmp;
			
		} else
			*s++ = *p++;
	}
	
	*s++ = '\0';
	return(TRUE);
	
}


static int trans_path(const char *str, char *buf)
{
	const char *p;
	char *s;
	int dot,done;
	
	
	for (p = str, s = buf; *p;) {
		if ((*p == '/' || *p == '\\')) {
			dot=done=0;
			
			*s++ = '/';
			p++;
			
			while(!done) {
				switch(*p) {
				case '\\':
				case '/':
					p++;
					dot=0;
				break;
				case '.':
					if (dot) {
						done=1;
						break;
					}
					
					if (*(p+1) != '.')
						p++;
					
					dot=1;
				break;
				default:
					done=1;
				break;
				}
			}
			
		} else
			*s++ = *p++;
	}
	
	*s++ = '\0';
	return(TRUE);
	
}


char *trans_uri(pool *p, const char *str, int flags)
{
	char buf[(strlen(str)+1)];
	char *ret;
	
	
	ret = ap_pstrdup(p,str);
	if (!ret)
		return(NULL);
	
	if (flags & TRANS_CHAR) {
		if (!trans_char(ret,buf))
			return(NULL);
	}
	
	strcpy(ret,buf);
	
	if (flags & TRANS_PATH) {		
		if (!trans_path(ret,buf))
			return(NULL);
	}
	
	strcpy(ret,buf);
	
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO,NULL,
			"trans_uri() = \"%s\"",ap_escape_logitem(p,buf));
	
	return (ret);
	
}


int check_uri(wasap_rec *wr, char *uri)
{
	char *ptr, *buf;
	unsigned int idx;
	
	
	if ((buf = trans_uri(wr->pool, uri, TRANS_CHAR)) == NULL) {
		LOG_ERROR(wr,APLOG_CRIT,
				"mod_wasap: trans_uri failed on uri: %s",
				ap_escape_logitem(wr->pool,uri));
		
		exec_alarm(wr->rec,"illegal uri");
		return (FALSE);
	}
	
	for (ptr = buf, idx = 0; *ptr && ((ptr - buf) < MAX_URILEN); ptr++) {
		
		if (!isascii(*ptr)) {
			
			LOG_ERROR(wr, APLOG_CRIT,
				"mod_wasap: found illegal char: [0x%02x]"
				" in URI[%s]",
				*ptr, ap_escape_logitem(wr->pool,buf));
			
			exec_alarm(wr->rec,"illegal character in uri");
			return (FALSE);
		}
		
		switch (*ptr) {
		
		case '.':
			if (*(ptr + 1) == '.')
				if (*(ptr - 1) == '/' || *(ptr + 2) == '/') {
					idx--;
					for (;
					     *ptr && (*(ptr + 1) == '.'
						      && *ptr != '/');
					     ptr++);
				}
			if (*(ptr + 1) == '/' && *(ptr - 1) == '/')
				ptr++;
			break;
		case '/':
			if (!idx || *(ptr - 1) != '/')
				idx++;
			break;

		}

		if (idx <= 0) {
			LOG_ERROR(wr, APLOG_CRIT,
				      "mod_wasap: reverse traversal: [%s]",
				      ap_escape_logitem(wr->pool,buf));
			
			exec_alarm(wr->rec,"reverse traversal");
			return (FALSE);
		}
		
	}
	
	if ((ptr - buf) > MAX_URILEN) {
		LOG_ERROR(wr, APLOG_CRIT,
			      "mod_wasap: URI is longer than MAX_URILEN[%u], [%s]",
			      MAX_URILEN, ap_escape_logitem(wr->pool,buf));
		
		exec_alarm(wr->rec,"uri is too long");
		return (FALSE);
	}
	
	return (TRUE);
	
}
